==> xkaapi-2.1.tar.gz: xkaapi-2.1/src/perfmodel/kaapi_perfmodel_task.c <==

#include <stdio.h>
#include <math.h>

#include "kaapi_impl.h"
#include "kaapi_tasklist.h"

#define KAAPI_PERFMODEL_TASK_FILE    "kaapi_perfmodel_task_dump"
#define KAAPI_PERFMODEL_MAX_PARAMS    8

/* time before task execution (temporary always) */
static double perfmodel_pre_task[KAAPI_MAX_PROCESSOR];

typedef struct kaapi_perfmodel_task_t {
  kaapi_format_id_t  fmtid;
  uint32_t key;
  struct perf {
    kaapi_lock_t lock;
    double mean;    /* mean of times */
    double sd;      /* standard deviation */
    double sum;     /* all sum of execution times */
    double sum_pow2;
    unsigned int samples; /* number of samples */
  } arch[KAAPI_PROC_TYPE_MAX];
} kaapi_perfmodel_task_t;

static kaapi_lock_t kaapi_perfmodel_hash_lock;
static kaapi_big_hashmap_t kaapi_perfmodel_hash;

/** return time in microseconds (us)
 */
static inline double kaapi_get_elapsedus(void)
{
  uint64_t retval;
  retval = kaapi_get_elapsedns();
  return (double)(retval / 1e3);
}

static inline uint8_t kaapi_get_format_id(const kaapi_format_id_t fmtid)
{
  /* see src/task/kaapi_fmt_resolvebyfmtid.c for explanation */
  uint8_t id = (uint8_t)(fmtid & (kaapi_format_id_t)0xFFUL);
  return id;
}

static inline FILE* kaapi_perfmodel_file_open_w(void)
{
  FILE* file = fopen(KAAPI_PERFMODEL_TASK_FILE, "w+");
  kaapi_assert(file != NULL);
  return file;
}

static inline FILE* kaapi_perfmodel_file_open_r(void)
{
  FILE* file = fopen(KAAPI_PERFMODEL_TASK_FILE, "r");
  return file;
}

static inline void kaapi_perfmodel_file_close(FILE* file)
{
  fclose(file);
}

static inline void kaapi_perfmodel_task_init_entry(kaapi_perfmodel_task_t* const perf,
                                                   const kaapi_format_id_t fmtid,
                                                   const uint32_t key)
{
  int j;
  perf->fmtid = fmtid;
  perf->key   = key;
  for(j= 0; j < KAAPI_PROC_TYPE_MAX; j++)
  {
    kaapi_atomic_initlock(&perf->arch[j].lock);
    perf->arch[j].mean     = 0.0f;
    perf->arch[j].sd       = 0.0f;
    perf->arch[j].sum      = 0.0f;
    perf->arch[j].sum_pow2 = 0.0f;
    perf->arch[j].samples  =  0;
  }
}

static inline void kaapi_perfmodel_task_destroy_entry(kaapi_perfmodel_task_t* const perf)
{
  int j;
#if defined(KAAPI_DEBUG)
  perf->fmtid = 0UL;
  perf->key   = 0UL;
#endif
  for(j= 0; j < KAAPI_PROC_TYPE_MAX; j++)
  {
    kaapi_atomic_destroylock(&perf->arch[j].lock);
#if defined(KAAPI_DEBUG)
    perf->arch[j].mean     = 0.0f;
    perf->arch[j].sd       = 0.0f;
    perf->arch[j].sum      = 0.0f;
    perf->arch[j].sum_pow2 = 0.0f;
    perf->arch[j].samples  =  0;
#endif
  }
}

static inline void kaapi_perfmodel_task_add_entry_(
                                                   uint32_t key,
                                                   kaapi_format_id_t fmtid,
                                                   int arch,
                                                   double mean,
                                                   double sd,
                                                   double sum,
                                                   double sum_pow2,
                                                   unsigned int samples
                                                   )
{  
  kaapi_hashentries_t *entry;
  kaapi_perfmodel_task_t* perfmodel_task;
  uint64_t hkey = (uint64_t)key;
  
  entry = kaapi_big_hashmap_findinsert(&kaapi_perfmodel_hash, (void*)hkey);
  if(entry->u.perf == 0)
  {
    entry->u.perf = (kaapi_perfmodel_task_t*)malloc(sizeof(kaapi_perfmodel_task_t));
    kaapi_perfmodel_task_init_entry(entry->u.perf, fmtid, key);
  }
  
  perfmodel_task = entry->u.perf;
  perfmodel_task->arch[arch].mean = mean;
  perfmodel_task->arch[arch].sd = sd;
  perfmodel_task->arch[arch].sum = sum;
  perfmodel_task->arch[arch].sum_pow2 = sum_pow2;
  perfmodel_task->arch[arch].samples = samples;
}

static inline void kaapi_perfmodel_file_parse(FILE* file)
{
  kaapi_format_id_t fmtid;
  uint32_t key;
  int arch, samples;
  double mean, sd, sum, sum_pow2;
  int ret;
  
  kaapi_atomic_lock(&kaapi_perfmodel_hash_lock);
  while(1)
  {
    ret = fscanf(file, "%x %u %d %le %le %le %le %d\n",
                 &key,
                 &fmtid,
                 &arch,
                 &mean,
                 &sd,
                 &sum,
                 &sum_pow2,
                 &samples
                 );
    if((ret == EOF) || (ret != KAAPI_PERFMODEL_MAX_PARAMS))
      break;
    
    kaapi_perfmodel_task_add_entry_(key, fmtid, arch, mean, sd, sum, sum_pow2, samples);
  }
  kaapi_atomic_unlock(&kaapi_perfmodel_hash_lock);
}

static inline void kaapi_perfmodel_file_dump(FILE* file)
{
  static const uint32_t map_size = KAAPI_HASHMAP_BIG_SIZE;
  kaapi_big_hashmap_t *const hmap = &kaapi_perfmodel_hash;
  kaapi_hashentries_t *entry;
  uint32_t i;
  int j;
  
  for (i = 0; i < map_size; ++i) {
    for (entry = hmap->entries[i]; entry != NULL; entry = entry->next) {
      kaapi_perfmodel_task_t* perfmodel_task = entry->u.perf;
      if (perfmodel_task == 0)
        continue;
      
      if(perfmodel_task->fmtid == 0)
        continue;
      
      for(j= 0; j < KAAPI_PROC_TYPE_MAX; j++)
      {
        if(perfmodel_task->arch[j].samples == 0)
          continue;
        
        fprintf(file, "%x %u %d %le %le %le %le %d\n",
                perfmodel_task->key,
                perfmodel_task->fmtid,
                j,
                perfmodel_task->arch[j].mean,
                perfmodel_task->arch[j].sd,
                perfmodel_task->arch[j].sum,
                perfmodel_task->arch[j].sum_pow2,
                perfmodel_task->arch[j].samples
                );        
      }
    }
  }  
}

static inline uint32_t kaapi_perfmodel_gen_key(kaapi_taskdescr_t* const td)
{
  unsigned int bytes = (unsigned int)kaapi_memory_taskdescr_affinity_get_size_bytes(td);
  uint32_t key;

  key = bytes ^ td->fmt->fmtid;
#if 0
  fprintf(stdout,"[%s] bytes=%x fmtid=%x bigkey=%x key=%x\n", __FUNCTION__,
          bytes,
          td->fmt->fmtid,
          bigkey,
          kaapi_hash_ulong(bigkey)
          );
  fflush(stdout);
#endif
  return key;
}

static inline void kaapi_perfmodel_task_add_entry(kaapi_taskdescr_t* const td)
{
  kaapi_hashentries_t *entry;
  uint32_t key;
  uint64_t hkey;
  
  key = kaapi_perfmodel_gen_key(td);
  hkey = (uint64_t)key;
  kaapi_atomic_lock(&kaapi_perfmodel_hash_lock);
  entry = kaapi_big_hashmap_findinsert(&kaapi_perfmodel_hash, (const void*)hkey);
  kaapi_atomic_unlock(&kaapi_perfmodel_hash_lock);
  if(entry->u.perf == 0)
  {
    entry->u.perf = (kaapi_perfmodel_task_t*)malloc(sizeof(kaapi_perfmodel_task_t));
    kaapi_perfmodel_task_init_entry(entry->u.perf, td->fmt->fmtid, key);
#if 0
    fprintf(stdout, "%d:%s: td=%p (%s) fmtid=%x\n", kaapi_get_self_kid(), __FUNCTION__,
            (void*)td, td->fmt->name, td->fmt->fmtid);
    fflush(stdout);
#endif
        
  }
}

static inline kaapi_perfmodel_task_t* kaapi_perfmodel_task_find_entry(kaapi_taskdescr_t* const td)
{
  kaapi_hashentries_t *entry;
  uint32_t key;
  uint64_t hkey;
  
  key = kaapi_perfmodel_gen_key(td);
  hkey = (uint64_t)key;
  entry = kaapi_big_hashmap_find(&kaapi_perfmodel_hash, (void*)hkey);
  if(entry == 0)
    return 0;
  
  return entry->u.perf;
}

static inline void kaapi_perfmodel_task_destroy_all_entry(void)
{
  static const uint32_t map_size = KAAPI_HASHMAP_BIG_SIZE;
  kaapi_big_hashmap_t *const hmap = &kaapi_perfmodel_hash;
  kaapi_hashentries_t *entry;
  uint32_t i;
  
  for (i = 0; i < map_size; ++i) {
    for (entry = hmap->entries[i]; entry != NULL; entry = entry->next) {
      kaapi_perfmodel_task_t* perf = entry->u.perf;
      if (perf == 0)
        continue;
      
      kaapi_perfmodel_task_destroy_entry(perf);
      free(perf);
    }
  }  
}

void kaapi_perfmodel_task_init(void)
{
  int i;
  FILE* file;
  
  for(i = 0; i < KAAPI_MAX_PROCESSOR; i++)
  {
    perfmodel_pre_task[i] = 0.0;
  }
  kaapi_big_hashmap_init(&kaapi_perfmodel_hash, 0);
  kaapi_atomic_initlock(&kaapi_perfmodel_hash_lock);
  
  file = kaapi_perfmodel_file_open_r();
  if(file != 0)
  {
    kaapi_perfmodel_file_parse(file);
    kaapi_perfmodel_file_close(file);
//    kaapi_perfmodel_print_stats();
  }
}

void kaapi_perfmodel_task_finalize(void)
{
//  kaapi_perfmodel_print_stats();
  FILE* file;
  
  if(kaapi_default_param.perfmodel_calibrate)
  {
    file = kaapi_perfmodel_file_open_w();
    if(file != 0)
    {
      kaapi_perfmodel_file_dump(file);
      kaapi_perfmodel_file_close(file);
    }
  }
  
  kaapi_perfmodel_task_destroy_all_entry();
  kaapi_big_hashmap_destroy(&kaapi_perfmodel_hash);
  kaapi_atomic_destroylock(&kaapi_perfmodel_hash_lock);
}

void kaapi_perfmodel_task_prologue(kaapi_processor_t* const kproc,
                                          kaapi_taskdescr_t* const td)
{
  if(kaapi_default_param.perfmodel_calibrate)
  {
    kaapi_perfmodel_task_add_entry(td);
    perfmodel_pre_task[kproc->kid] = kaapi_get_elapsedus(); /* time now */    
  }
}

/* Call after task execution to measure execution time */
void kaapi_perfmodel_task_epilogue(kaapi_processor_t* const kproc,
                                          kaapi_taskdescr_t* const td)
{
  if(kaapi_default_param.perfmodel_calibrate)
  {
    double t1 = kaapi_get_elapsedus();
    double tdelta = t1 - perfmodel_pre_task[kproc->kid];
    const int arch = kaapi_processor_get_type(kproc);
    
    kaapi_perfmodel_task_t* perfmodel_task = kaapi_perfmodel_task_find_entry(td);

    kaapi_atomic_lock(&perfmodel_task->arch[arch].lock);
    perfmodel_task->arch[arch].samples++;
    perfmodel_task->arch[arch].sum += tdelta;
    perfmodel_task->arch[arch].sum_pow2 += (tdelta*tdelta);
    
    unsigned int n = perfmodel_task->arch[arch].samples;
    perfmodel_task->arch[arch].mean = perfmodel_task->arch[arch].sum / n;
    perfmodel_task->arch[arch].sd =
      sqrt( (perfmodel_task->arch[arch].sum_pow2 - (perfmodel_task->arch[arch].sum*perfmodel_task->arch[arch].sum)/n)/n );
    kaapi_atomic_unlock(&perfmodel_task->arch[arch].lock);
  }
}

void kaapi_perfmodel_print_stats(void)
{
#if defined(KAAPI_DEBUG)    
  static const uint32_t map_size = KAAPI_HASHMAP_BIG_SIZE;
  kaapi_big_hashmap_t *const hmap = &kaapi_perfmodel_hash;
  kaapi_hashentries_t *entry;
  uint32_t i;
  int j;
  
  for (i = 0; i < map_size; ++i) {
    for (entry = hmap->entries[i]; entry != NULL; entry = entry->next) {
      kaapi_perfmodel_task_t* perfmodel_task = entry->u.perf;
      if (perfmodel_task == 0)
        continue;
      
      if(perfmodel_task->fmtid == 0)
        continue;
      
      for(j= 0; j < KAAPI_PROC_TYPE_MAX; j++)
      {
        if(perfmodel_task->arch[j].samples == 0)
          continue;
        
        fprintf(stdout, "[%s] (%s) key=%x fmtid=%u samples=%d mean=%.10f sd=%.10f\n",
                __FUNCTION__,
                (j == KAAPI_PROC_TYPE_CPU)?"CPU":"CUDA",
                perfmodel_task->key,
                perfmodel_task->fmtid,
                perfmodel_task->arch[j].samples,
                perfmodel_task->arch[j].mean,
                perfmodel_task->arch[j].sd
                );
        fflush(stdout);
      }
      
      if((perfmodel_task->arch[KAAPI_PROC_TYPE_CPU].samples != 0) && (perfmodel_task->arch[KAAPI_PROC_TYPE_CUDA].samples != 0))
      {
        fprintf(stdout, "[%s] key=%x fmtid=%u speedup=%.10f\n",
                __FUNCTION__,
                perfmodel_task->key,
                perfmodel_task->fmtid,
                (perfmodel_task->arch[KAAPI_PROC_TYPE_CPU].mean/perfmodel_task->arch[KAAPI_PROC_TYPE_CUDA].mean)
                );
        fflush(stdout);        
      }
    }
  }
#endif
}

double kaapi_perfmodel_task_predict(kaapi_processor_t* const kproc,
                                         kaapi_taskdescr_t* const td)
{
  const int arch = kaapi_processor_get_type(kproc);
  double predicted= 0.f;
  
  if(td->fmt == 0)
    return 0.f;
  
  kaapi_perfmodel_task_t* const perfmodel_task = kaapi_perfmodel_task_find_entry(td);
  
  if(perfmodel_task == 0)
    return 0.f;
  
  kaapi_atomic_lock(&perfmodel_task->arch[arch].lock);
  predicted = perfmodel_task->arch[arch].mean;
  kaapi_atomic_unlock(&perfmodel_task->arch[arch].lock);
  
  return predicted;
}

double kaapi_perfmodel_task_predict_by_arch(const int arch, kaapi_taskdescr_t* const td)
{
  double predicted= 0.f;
  
  if(td->fmt == 0)
    return 0.f;
  
  kaapi_perfmodel_task_t* const perfmodel_task = kaapi_perfmodel_task_find_entry(td);
  
  if(perfmodel_task == 0)
    return 0.f;
  
  kaapi_atomic_lock(&perfmodel_task->arch[arch].lock);
  predicted = perfmodel_task->arch[arch].mean;
  kaapi_atomic_unlock(&perfmodel_task->arch[arch].lock);
  
  return predicted;
}

==> xkaapi-2.2.tar.gz: xkaapi-2.2/src/perfmodel/kaapi_perfmodel_task.c <==

#include <stdio.h>
#include <math.h>

#include "kaapi_impl.h"
#include "kaapi_tasklist.h"

#define KAAPI_PERFMODEL_TASK_FILE    "kaapi_perfmodel_task_dump"
#define KAAPI_PERFMODEL_MAX_PARAMS    10

#define KAAPI_PERFMODEL_MIN_SAMPLES   30

/* time before task execution (temporary always) */
static double perfmodel_pre_task[KAAPI_MAX_PROCESSOR];

typedef struct kaapi_perfmodel_task_t {
  kaapi_format_id_t  fmtid;
  uint32_t key;
  unsigned int bytes;
  struct perf {
    kaapi_lock_t lock;
    double mean;    /* mean of times */
    double sd;      /* standard deviation */
    double sum;     /* all sum of execution times */
    double sum_pow2;
    unsigned int samples; /* number of samples */
  } arch[KAAPI_PROC_TYPE_MAX];
} kaapi_perfmodel_task_t;

static kaapi_lock_t kaapi_perfmodel_hash_lock;
static kaapi_big_hashmap_t kaapi_perfmodel_hash;

/** return time in microseconds (us)
 */
static inline double kaapi_get_elapsedus(void)
{
  uint64_t retval;
  retval = kaapi_get_elapsedns();
  return (double)(retval / 1e3);
}

static inline uint8_t kaapi_get_format_id(const kaapi_format_id_t fmtid)
{
  /* see src/task/kaapi_fmt_resolvebyfmtid.c for explanation */
  uint8_t id = (uint8_t)(fmtid & (kaapi_format_id_t)0xFFUL);
  return id;
}

static inline FILE* kaapi_perfmodel_file_open_w(void)
{
  FILE* file = fopen(KAAPI_PERFMODEL_TASK_FILE, "w+");
  kaapi_assert(file != NULL);
  return file;
}

static inline FILE* kaapi_perfmodel_file_open_r(void)
{
  FILE* file = fopen(KAAPI_PERFMODEL_TASK_FILE, "r");
  return file;
}

static inline void kaapi_perfmodel_file_close(FILE* file)
{
  fclose(file);
}

static inline void kaapi_perfmodel_task_init_entry(kaapi_perfmodel_task_t* const perf,
                                                   const kaapi_format_id_t fmtid,
                                                   const uint32_t key,
                                                   const unsigned int bytes)
{
  int j;
  perf->fmtid = fmtid;
  perf->key   = key;
  perf->bytes = bytes;
  for(j= 0; j < KAAPI_PROC_TYPE_MAX; j++)
  {
    kaapi_atomic_initlock(&perf->arch[j].lock);
    perf->arch[j].mean     = 0.0f;
    perf->arch[j].sd       = 0.0f;
    perf->arch[j].sum      = 0.0f;
    perf->arch[j].sum_pow2 = 0.0f;
    perf->arch[j].samples  =  0;
  }
}

static inline void kaapi_perfmodel_task_destroy_entry(kaapi_perfmodel_task_t* const perf)
{
  int j;
#if defined(KAAPI_DEBUG)
  perf->fmtid = 0UL;
  perf->key   = 0UL;
  perf->bytes = 0UL;
#endif
  for(j= 0; j < KAAPI_PROC_TYPE_MAX; j++)
  {
    kaapi_atomic_destroylock(&perf->arch[j].lock);
#if defined(KAAPI_DEBUG)
    perf->arch[j].mean     = 0.0f;
    perf->arch[j].sd       = 0.0f;
    perf->arch[j].sum      = 0.0f;
    perf->arch[j].sum_pow2 = 0.0f;
    perf->arch[j].samples  =  0;
#endif
  }
}

static inline void kaapi_perfmodel_task_add_entry_(
                                                   const unsigned int bytes,
                                                   uint32_t key,
                                                   kaapi_format_id_t fmtid,
                                                   int arch,
                                                   double mean,
                                                   double sd,
                                                   double sum,
                                                   double sum_pow2,
                                                   unsigned int samples
                                                   )
{  
  kaapi_hashentries_t *entry;
  kaapi_perfmodel_task_t* perfmodel_task;
  uint64_t hkey = (uint64_t)key;
  
  entry = kaapi_big_hashmap_findinsert(&kaapi_perfmodel_hash, (void*)hkey);
  if(entry->u.perf == 0)
  {
    entry->u.perf = malloc(sizeof(kaapi_perfmodel_task_t));
    kaapi_perfmodel_task_init_entry((kaapi_perfmodel_task_t*)entry->u.perf, fmtid, key, bytes);
  }
  
  perfmodel_task = (kaapi_perfmodel_task_t*)entry->u.perf;
  perfmodel_task->arch[arch].mean = mean;
  perfmodel_task->arch[arch].sd = sd;
  perfmodel_task->arch[arch].sum = sum;
  perfmodel_task->arch[arch].sum_pow2 = sum_pow2;
  perfmodel_task->arch[arch].samples = samples;  
}

static inline char* kaapi_get_arch_string(int arch)
{
  switch (arch) {
    case KAAPI_PROC_TYPE_CPU:
      return "CPU";
      break;
      
    case KAAPI_PROC_TYPE_CUDA:
      return "CUDA";
      
    case KAAPI_PROC_TYPE_MPSOC:
      return "MPSOC";
      
    default:
      break;
  }
  
  return "none";
}

static inline int kaapi_get_string_arch(char* const str)
{
  if(strcmp(str, "CPU") == 0)
    return KAAPI_PROC_TYPE_CPU;
  
  if(strcmp(str, "CUDA") == 0)
    return KAAPI_PROC_TYPE_CUDA;
  
  if(strcmp(str, "MPSOC") == 0)
    return KAAPI_PROC_TYPE_MPSOC;
  
  return KAAPI_PROC_TYPE_CPU;
}

static inline void kaapi_perfmodel_file_parse(FILE* file)
{
  kaapi_format_id_t fmtid;
  uint32_t key;
  int arch, samples;
  double mean, sd, sum, sum_pow2;
  int ret;
  char arch_str[10];
  char name[30];
  unsigned int size = 0;
  
  kaapi_atomic_lock(&kaapi_perfmodel_hash_lock);
  while(1)
  {
    ret = fscanf(file, "%s %u %x %u %s %le %le %le %le %d\n",
                 name,
                 &size,
                 &key,
                 &fmtid,
                 arch_str,
                 &mean,
                 &sd,
                 &sum,
                 &sum_pow2,
                 &samples
                 );
    if((ret == EOF) || (ret != KAAPI_PERFMODEL_MAX_PARAMS))
      break;
    
    arch = kaapi_get_string_arch(arch_str);
    kaapi_perfmodel_task_add_entry_(size, key, fmtid, arch, mean, sd, sum, sum_pow2, samples);
    if(samples < KAAPI_PERFMODEL_MIN_SAMPLES)
    {
      fprintf(stdout, "*** Kaapi: WARNING samples from task '%s' are insufficient for performance model (samples are %d).\n",
              name, samples);
      fflush(stdout);
    }
  }
  kaapi_atomic_unlock(&kaapi_perfmodel_hash_lock);
}

static inline void kaapi_perfmodel_file_dump(FILE* file)
{
  static const uint32_t map_size = KAAPI_HASHMAP_BIG_SIZE;
  kaapi_big_hashmap_t *const hmap = &kaapi_perfmodel_hash;
  kaapi_hashentries_t *entry;
  uint32_t i;
  int j;
  char name[30];
  
  for (i = 0; i < map_size; ++i) {
    for (entry = hmap->entries[i]; entry != NULL; entry = entry->next) {
      kaapi_perfmodel_task_t* perfmodel_task = entry->u.perf;
      if (perfmodel_task == 0)
        continue;
      
      if(perfmodel_task->fmtid == 0)
        continue;
      
      kaapi_format_t* const fmt = kaapi_format_resolvebyfmit(perfmodel_task->fmtid);
      if(fmt != 0)
        snprintf(name, 30, "%s", fmt->name);
      else
        snprintf(name, 30, "none");
      
      for(j= 0; j < KAAPI_PROC_TYPE_MAX; j++)
      {
        if(perfmodel_task->arch[j].samples == 0)
          continue;
        
        fprintf(file, "%s %u %x %u %s %le %le %le %le %d\n",
                name,
                perfmodel_task->bytes,
                perfmodel_task->key,
                perfmodel_task->fmtid,
                kaapi_get_arch_string(j),
                perfmodel_task->arch[j].mean,
                perfmodel_task->arch[j].sd,
                perfmodel_task->arch[j].sum,
                perfmodel_task->arch[j].sum_pow2,
                perfmodel_task->arch[j].samples
                );        
      }
    }
  }  
}

static inline uint32_t kaapi_perfmodel_gen_key(kaapi_taskdescr_t* const td, const unsigned int bytes)
{
  uint32_t key;

  key = bytes ^ td->fmt->fmtid;
#if 0
  fprintf(stdout,"[%s] bytes=%x fmtid=%x bigkey=%x key=%x\n", __FUNCTION__,
          bytes,
          td->fmt->fmtid,
          bigkey,
          kaapi_hash_ulong(bigkey)
          );
  fflush(stdout);
#endif
  return key;
}

static inline void kaapi_perfmodel_task_add_entry(kaapi_taskdescr_t* const td)
{
  kaapi_hashentries_t *entry;
  uint32_t key;
  uint64_t hkey;
  unsigned int bytes = (unsigned int)kaapi_memory_taskdescr_get_footprint(td);
  
  key = kaapi_perfmodel_gen_key(td, bytes);
  hkey = (uint64_t)key;
  kaapi_atomic_lock(&kaapi_perfmodel_hash_lock);
  entry = kaapi_big_hashmap_findinsert(&kaapi_perfmodel_hash, (const void*)hkey);
  kaapi_atomic_unlock(&kaapi_perfmodel_hash_lock);
  if(entry->u.perf == 0)
  {
    entry->u.perf = (kaapi_perfmodel_task_t*)malloc(sizeof(kaapi_perfmodel_task_t));
    kaapi_perfmodel_task_init_entry(entry->u.perf, td->fmt->fmtid, key, bytes);
#if 0
    fprintf(stdout, "%d:%s: td=%p (%s) fmtid=%x\n", kaapi_get_self_kid(), __FUNCTION__,
            (void*)td, td->fmt->name, td->fmt->fmtid);
    fflush(stdout);
#endif
        
  }
}

static inline kaapi_perfmodel_task_t* kaapi_perfmodel_task_find_entry(kaapi_taskdescr_t* const td)
{
  kaapi_hashentries_t *entry;
  uint32_t key;
  uint64_t hkey;
  unsigned int bytes = (unsigned int)kaapi_memory_taskdescr_get_footprint(td);  
  
  key = kaapi_perfmodel_gen_key(td, bytes);
  hkey = (uint64_t)key;
  entry = kaapi_big_hashmap_find(&kaapi_perfmodel_hash, (void*)hkey);
  if(entry == 0)
    return 0;
  
  return entry->u.perf;
}

static inline void kaapi_perfmodel_task_destroy_all_entry(void)
{
  static const uint32_t map_size = KAAPI_HASHMAP_BIG_SIZE;
  kaapi_big_hashmap_t *const hmap = &kaapi_perfmodel_hash;
  kaapi_hashentries_t *entry;
  uint32_t i;
  
  for (i = 0; i < map_size; ++i) {
    for (entry = hmap->entries[i]; entry != NULL; entry = entry->next) {
      kaapi_perfmodel_task_t* perf = entry->u.perf;
      if (perf == 0)
        continue;
      
      kaapi_perfmodel_task_destroy_entry(perf);
      free(perf);
    }
  }  
}

void kaapi_perfmodel_task_init(void)
{
  int i;
  FILE* file;
  
  for(i = 0; i < KAAPI_MAX_PROCESSOR; i++)
  {
    perfmodel_pre_task[i] = 0.0;
  }
  kaapi_big_hashmap_init(&kaapi_perfmodel_hash, 0);
  kaapi_atomic_initlock(&kaapi_perfmodel_hash_lock);
  
  file = kaapi_perfmodel_file_open_r();
  if(file != 0)
  {
    kaapi_perfmodel_file_parse(file);
    kaapi_perfmodel_file_close(file);
//    kaapi_perfmodel_print_stats();
  }
}

void kaapi_perfmodel_task_finalize(void)
{
//  kaapi_perfmodel_print_stats();
  FILE* file;
  
  if(kaapi_default_param.perfmodel_calibrate)
  {
    file = kaapi_perfmodel_file_open_w();
    if(file != 0)
    {
      kaapi_perfmodel_file_dump(file);
      kaapi_perfmodel_file_close(file);
    }
  }
  
  kaapi_perfmodel_task_destroy_all_entry();
  kaapi_big_hashmap_destroy(&kaapi_perfmodel_hash);
  kaapi_atomic_destroylock(&kaapi_perfmodel_hash_lock);
}

void kaapi_perfmodel_task_prologue(kaapi_processor_t* const kproc,
                                          kaapi_taskdescr_t* const td)
{
  if(kaapi_default_param.perfmodel_calibrate)
  {
    kaapi_perfmodel_task_add_entry(td);
    perfmodel_pre_task[kproc->kid] = kaapi_get_elapsedus(); /* time now */    
  }
}

/* Call after task execution to measure execution time */
void kaapi_perfmodel_task_epilogue(kaapi_processor_t* const kproc,
                                          kaapi_taskdescr_t* const td)
{
  if(kaapi_default_param.perfmodel_calibrate)
  {
    double t1 = kaapi_get_elapsedus();
    double tdelta = t1 - perfmodel_pre_task[kproc->kid];
    const int arch = kaapi_processor_get_type(kproc);
    
    kaapi_perfmodel_task_t* perfmodel_task = kaapi_perfmodel_task_find_entry(td);

    kaapi_atomic_lock(&perfmodel_task->arch[arch].lock);
    perfmodel_task->arch[arch].samples++;
    perfmodel_task->arch[arch].sum += tdelta;
    perfmodel_task->arch[arch].sum_pow2 += (tdelta*tdelta);
    
    unsigned int n = perfmodel_task->arch[arch].samples;
    perfmodel_task->arch[arch].mean = perfmodel_task->arch[arch].sum / n;
    perfmodel_task->arch[arch].sd =
      sqrt( (perfmodel_task->arch[arch].sum_pow2 - (perfmodel_task->arch[arch].sum*perfmodel_task->arch[arch].sum)/n)/n );
    kaapi_atomic_unlock(&perfmodel_task->arch[arch].lock);
  }
}

void kaapi_perfmodel_print_stats(void)
{
#if defined(KAAPI_DEBUG)    
  static const uint32_t map_size = KAAPI_HASHMAP_BIG_SIZE;
  kaapi_big_hashmap_t *const hmap = &kaapi_perfmodel_hash;
  kaapi_hashentries_t *entry;
  uint32_t i;
  int j;
  
  for (i = 0; i < map_size; ++i) {
    for (entry = hmap->entries[i]; entry != NULL; entry = entry->next) {
      kaapi_perfmodel_task_t* perfmodel_task = entry->u.perf;
      if (perfmodel_task == 0)
        continue;
      
      if(perfmodel_task->fmtid == 0)
        continue;

      kaapi_format_t* const fmt = kaapi_format_resolvebyfmit(perfmodel_task->fmtid);
      char* name;
      if(fmt != 0)
        name = fmt->name;
      else
        name = "";
      
      for(j= 0; j < KAAPI_PROC_TYPE_MAX; j++)
      {
        if(perfmodel_task->arch[j].samples == 0)
          continue;
        
        fprintf(stdout, "[%s] (%s) key=%x fmtid=%u (%s) samples=%d mean=%.10f sd=%.10f\n",
                __FUNCTION__,
                (j == KAAPI_PROC_TYPE_CPU)?"CPU":"CUDA",
                perfmodel_task->key,
                perfmodel_task->fmtid,
                name,
                perfmodel_task->arch[j].samples,
                perfmodel_task->arch[j].mean,
                perfmodel_task->arch[j].sd
                );
        fflush(stdout);
      }
      
      if((perfmodel_task->arch[KAAPI_PROC_TYPE_CPU].samples != 0) && (perfmodel_task->arch[KAAPI_PROC_TYPE_CUDA].samples != 0))
      {
        fprintf(stdout, "[%s] key=%x fmtid=%u speedup=%.10f\n",
                __FUNCTION__,
                perfmodel_task->key,
                perfmodel_task->fmtid,
                (perfmodel_task->arch[KAAPI_PROC_TYPE_CPU].mean/perfmodel_task->arch[KAAPI_PROC_TYPE_CUDA].mean)
                );
        fflush(stdout);        
      }
    }
  }
#endif
}

double kaapi_perfmodel_task_predict(kaapi_processor_t* const kproc,
                                         kaapi_taskdescr_t* const td)
{
  const int arch = kaapi_processor_get_type(kproc);
  double predicted= 0.f;
  
  if(td->fmt == 0)
    return 0.f;
  
  kaapi_perfmodel_task_t* const perfmodel_task = kaapi_perfmodel_task_find_entry(td);
  
  if(perfmodel_task == 0)
    return 0.f;
  
  predicted = perfmodel_task->arch[arch].mean;
  
  return predicted;
}

double kaapi_perfmodel_task_predict_by_arch(const int arch, kaapi_taskdescr_t* const td)
{
  double predicted= 0.f;
  
  if(td->fmt == 0)
    return 0.f;
  
  kaapi_perfmodel_task_t* const perfmodel_task = kaapi_perfmodel_task_find_entry(td);
  
  if(perfmodel_task == 0)
    return 0.f;
  
  predicted = perfmodel_task->arch[arch].mean;
  
  return predicted;
}
